/**
* @file modules/generated/module_list.h
* @brief File containing includes of generated module files. Generated by json_to_array.py
* @copyright (c) 2021 Avast Software, licensed under the MIT license
*/
#pragma once

#include <vector>

#include "yaramod/types/modules/module_content.h"
#include "module_magic_generated.h"
#include "module_pe_generated.h"
#include "module_math_generated.h"
#include "module_console_generated.h"
#include "module_dotnet_generated.h"
#include "module_elf_generated.h"
#include "module_time_generated.h"
#include "module_cuckoo_deprecated_generated.h"
#include "module_cuckoo_generated.h"
#include "module_macho_generated.h"
#include "module_dex_generated.h"
#include "module_hash_generated.h"


namespace yaramod {

namespace modules {

class ModuleList {
public:
	std::vector<ModuleContent> list = {Gmodule_magic{}, Gmodule_pe{}, Gmodule_math{}, Gmodule_console{}, Gmodule_dotnet{}, Gmodule_elf{}, Gmodule_time{}, Gmodule_cuckoo_deprecated{}, Gmodule_cuckoo{}, Gmodule_macho{}, Gmodule_dex{}, Gmodule_hash{}};
};

} //namespace modules

} // namespace yaramod
